from __future__ import annotations

import secrets
from collections.abc import Sequence
from datetime import datetime, timezone
from typing import overload
from urllib.parse import urlparse
from uuid import uuid4

import requests
from cognite.client import CogniteClient
from cognite.client.data_classes import UserProfile, filters
from cognite.client.exceptions import CogniteAPIError

from .shop_run import SHOPRun, ShopRunEvent, SHOPRunList

DEFAULT_READ_LIMIT = 25


def _now_isoformat() -> str:
    """Current time in ISO format"""
    return datetime.now(timezone.utc).isoformat().replace("+00:00", "Z")


def _unique_short_str(nbytes: int) -> str:
    return secrets.token_hex(nbytes=nbytes)


class SHOPRunAPI:
    def __init__(self, client: CogniteClient, dataset_id: int, cogshop_version: str):
        self._cdf = client
        self._dataset_id = dataset_id
        self.cogshop_version = cogshop_version

    def trigger_case(self, case_file: str, watercourse: str, source: str | None = None) -> SHOPRun:
        """
        Trigger SHOP for a given case file.

        Args:
            case_file: Case file as a string. Expected to be YAML.
            watercourse: The watercourse to run SHOP this case file is for.
            source: The source of the SHOP trigger. If nothing is passed, the method will
                try to detect the service principal of the current user.

        Returns:
            The new SHOP run created.
        """
        user: UserProfile | None
        try:
            user = self._cdf.iam.user_profiles.me()
        except CogniteAPIError as e:
            if e.code == 404:
                user = None
            else:
                raise

        if source is None and user:
            source = user.user_identifier
        elif source is None:
            source = "manual"

        if user:
            display_name = user.display_name
        else:
            display_name = "Unknown"

        meta = self._cdf.files.upload_bytes(
            content=case_file,
            name=f"Manual Case by {display_name}",
            mime_type="application/yaml",
            external_id=f"cog_shop_manual/{uuid4()!s}",
            metadata={"shop:type": "cog_shop_case", "user": display_name},
            data_set_id=self._dataset_id,
            source=source,
        )
        now = datetime.now(timezone.utc)
        now_isoformat = now.isoformat().replace("+00:00", "Z")

        new_event = SHOPRun(
            external_id=f"SHOP_RUN_{now_isoformat}_{_unique_short_str(3)}",
            watercourse=watercourse,
            start=now,
            end=None,
            _case_file_external_id=meta.external_id,
            _shop_files=[],
            shop_version=self.cogshop_version,
            _client=self._cdf,
            source=source,
        )
        self._cdf.events.create(new_event.as_cdf_event(self._dataset_id))

        self._trigger_shop_container(new_event)
        return new_event

    def _trigger_shop_container(self, event: SHOPRun):
        def auth(r: requests.PreparedRequest) -> requests.PreparedRequest:
            auth_header_name, auth_header_value = self._cdf._config.credentials.authorization_header()
            r.headers[auth_header_name] = auth_header_value
            return r

        response = requests.post(
            url=self._shop_url(),
            json={"shopEventExternalId": event.external_id, "cogShopVersion": self.cogshop_version},
            auth=auth,
        )
        response.raise_for_status()

    def _shop_url(self) -> str:
        project = self._cdf.config.project

        cluster = urlparse(self._cdf.config.base_url).netloc.split(".", 1)[0]

        environment = project.split("-")[-1]
        if environment in {"dev", "staging"}:
            stage = ".staging"
        elif environment == "prod":
            stage = ""
        else:
            raise ValueError(f"Can't detect prod/staging from project name: {project!r}")

        return f"https://power-ops-api{stage}.{cluster}.cognite.ai/{project}/run-shop"

    def list(self, watercourse: str | list[str] | None = None, limit: int = DEFAULT_READ_LIMIT) -> SHOPRunList:
        """List the filtered SHOP runs.

        Args:
            watercourse: The watercourse to filter on.
            limit: The maximum number of SHOP runs to return. D

        Returns:
            A list of SHOP runs.
        """

        is_type = filters.Equals("type", ShopRunEvent.event_type)
        extra_filters = []
        if watercourse:
            watercourses = watercourse if isinstance(watercourse, list) else [watercourse]
            is_watercourse = filters.ContainsAny(["metadata", ShopRunEvent.watercourse], [watercourses])
            extra_filters.append(is_watercourse)

        selected = filters.And(is_type, *extra_filters)
        events = self._cdf.events.filter(selected, limit=limit)

        return SHOPRunList.load(events)

    @overload
    def retrieve(self, external_id: str) -> SHOPRun | None:
        ...

    @overload
    def retrieve(self, external_id: Sequence[str]) -> SHOPRunList:
        ...

    def retrieve(
        self, external_id: str | Sequence[str], ignore_unknown_ids: bool = True
    ) -> SHOPRun | SHOPRunList | None:
        """
        Retrieve a SHOP run.

        Args:
            external_id: The external id(s) of the SHOP run(s) to retrieve.
            ignore_unknown_ids: Whether to ignore unknown ids or raise an error.

        Returns:
            The SHOP run(s). None if it is a single external id, and it is not found.
        """
        if isinstance(external_id, str):
            event = self._cdf.events.retrieve(external_id=external_id)
            if event is None:
                return None
            return SHOPRun.load(event)
        elif isinstance(external_id, Sequence):
            return SHOPRunList.load(
                self._cdf.events.retrieve_multiple(external_ids=external_id, ignore_unknown_ids=ignore_unknown_ids)
            )
        else:
            raise TypeError(f"Invalid type {type(external_id)} for external_id.")



"""
This package contains the v2 version of the models in the resync package. All of these models
are based on Data Modeling in CDF.

* ProductionModel - This is the data model version of the ProductionModel in v1.
* Dayahead - This is the data model version of the Dayahead part of the MarketModel in v1.
* RKOM - This is the data model version of the RKOM part of the MarketModel in v1.
* Benchmarking - This is the data model version of the Benchmarking part of the MarketModel in v1.
* CogSHOP - This is the updated version of the CogShop1Asset in v1.
"""

from .cogshop import CogShopDataModel
from .graphql_schemas import GRAPHQL_MODELS
from .market_dm import BenchmarkMarketDataModel, DayAheadMarketDataModel, RKOMMarketDataModel
from .production_dm import ProductionModelDM

__all__ = [
    "GRAPHQL_MODELS",
    "CogShopDataModel",
    "ProductionModelDM",
    "DayAheadMarketDataModel",
    "RKOMMarketDataModel",
    "BenchmarkMarketDataModel",
]



from __future__ import annotations

from pydantic import Field

from cognite.powerops.client.data_classes import (
    OutputContainerApply,
    ScenarioMappingApply,
    ScenarioTemplateApply,
    ValueTransformationApply,
)
from cognite.powerops.resync.models._shared_v1_v2.cogshop_model import CogShopCore, ExternalID
from cognite.powerops.resync.models.base import DataModel


class CogShopDataModel(CogShopCore, DataModel):
    scenario_templates: list[ScenarioTemplateApply] = Field(default_factory=list)
    base_mappings: dict[ExternalID, ScenarioMappingApply] = Field(default_factory=dict)
    output_definitions: dict[ExternalID, OutputContainerApply] = Field(default_factory=dict)
    value_transformations: dict[ExternalID, ValueTransformationApply] = Field(default_factory=dict)



from __future__ import annotations

from typing import ClassVar, Optional

from cognite.client import data_modeling as dm
from pydantic import Field

from cognite.powerops.client._generated.data_classes._core import DomainModel, DomainModelApply, TypeList

__all__ = ["NordPoolMarket", "NordPoolMarketApply", "NordPoolMarketList"]


class NordPoolMarket(DomainModel):
    space: ClassVar[str] = "power-ops"
    max_price: Optional[float] = Field(None, alias="maxPrice")
    min_price: Optional[float] = Field(None, alias="minPrice")
    name: Optional[str] = None
    price_steps: Optional[int] = Field(None, alias="priceSteps")
    price_unit: Optional[str] = Field(None, alias="priceUnit")
    tick_size: Optional[float] = Field(None, alias="tickSize")
    time_unit: Optional[str] = Field(None, alias="timeUnit")
    timezone: Optional[str] = None
    trade_lot: Optional[float] = Field(None, alias="tradeLot")


class NordPoolMarketApply(DomainModelApply):
    space: ClassVar[str] = "power-ops"
    max_price: Optional[float] = None
    min_price: Optional[float] = None
    name: Optional[str] = None
    price_steps: Optional[int] = None
    price_unit: Optional[str] = None
    tick_size: Optional[float] = None
    time_unit: Optional[str] = None
    timezone: Optional[str] = None
    trade_lot: Optional[float] = None

    def _to_instances_apply(self, cache: set[str]) -> dm.InstancesApply:
        if self.external_id in cache:
            return dm.InstancesApply(dm.NodeApplyList([]), dm.EdgeApplyList([]))

        sources = []
        properties = {}
        if self.max_price is not None:
            properties["maxPrice"] = self.max_price
        if self.min_price is not None:
            properties["minPrice"] = self.min_price
        if self.name is not None:
            properties["name"] = self.name
        if self.price_steps is not None:
            properties["priceSteps"] = self.price_steps
        if self.price_unit is not None:
            properties["priceUnit"] = self.price_unit
        if self.tick_size is not None:
            properties["tickSize"] = self.tick_size
        if self.time_unit is not None:
            properties["timeUnit"] = self.time_unit
        if self.timezone is not None:
            properties["timezone"] = self.timezone
        if self.trade_lot is not None:
            properties["tradeLot"] = self.trade_lot
        if properties:
            source = dm.NodeOrEdgeData(
                source=dm.ContainerId("power-ops", "NordPoolMarket"),
                properties=properties,
            )
            sources.append(source)
        if sources:
            this_node = dm.NodeApply(
                space=self.space,
                external_id=self.external_id,
                existing_version=self.existing_version,
                sources=sources,
            )
            nodes = [this_node]
        else:
            nodes = []

        edges = []
        cache.add(self.external_id)

        return dm.InstancesApply(dm.NodeApplyList(nodes), dm.EdgeApplyList(edges))


class NordPoolMarketList(TypeList[NordPoolMarket]):
    _NODE = NordPoolMarket



from __future__ import annotations

from typing import TYPE_CHECKING, ClassVar, Optional, Union

from cognite.client import data_modeling as dm
from pydantic import Field

from cognite.powerops.client._generated.data_classes._core import DomainModel, DomainModelApply, TypeList

if TYPE_CHECKING:
    from cognite.powerops.client._generated.data_classes._command_configs import CommandConfigApply
    from cognite.powerops.client._generated.data_classes._scenario_mappings import ScenarioMappingApply
    from cognite.powerops.client._generated.data_classes._scenario_templates import ScenarioTemplateApply

__all__ = ["Scenario", "ScenarioApply", "ScenarioList"]


class Scenario(DomainModel):
    space: ClassVar[str] = "power-ops"
    commands: Optional[str] = None
    mapping: Optional[str] = None
    name: Optional[str] = None
    template: Optional[str] = None


class ScenarioApply(DomainModelApply):
    space: ClassVar[str] = "power-ops"
    commands: Optional[Union[CommandConfigApply, str]] = Field(None, repr=False)
    mapping: Optional[Union[ScenarioMappingApply, str]] = Field(None, repr=False)
    name: Optional[str] = None
    template: Optional[Union[ScenarioTemplateApply, str]] = Field(None, repr=False)

    def _to_instances_apply(self, cache: set[str]) -> dm.InstancesApply:
        if self.external_id in cache:
            return dm.InstancesApply(dm.NodeApplyList([]), dm.EdgeApplyList([]))

        sources = []
        properties = {}
        if self.commands is not None:
            properties["commands"] = {
                "space": "power-ops",
                "externalId": self.commands if isinstance(self.commands, str) else self.commands.external_id,
            }
        if self.mapping is not None:
            properties["mapping"] = {
                "space": "power-ops",
                "externalId": self.mapping if isinstance(self.mapping, str) else self.mapping.external_id,
            }
        if self.name is not None:
            properties["name"] = self.name
        if self.template is not None:
            properties["template"] = {
                "space": "power-ops",
                "externalId": self.template if isinstance(self.template, str) else self.template.external_id,
            }
        if properties:
            source = dm.NodeOrEdgeData(
                source=dm.ContainerId("power-ops", "Scenario"),
                properties=properties,
            )
            sources.append(source)
        if sources:
            this_node = dm.NodeApply(
                space=self.space,
                external_id=self.external_id,
                existing_version=self.existing_version,
                sources=sources,
            )
            nodes = [this_node]
        else:
            nodes = []

        edges = []
        cache.add(self.external_id)

        if isinstance(self.commands, DomainModelApply):
            instances = self.commands._to_instances_apply(cache)
            nodes.extend(instances.nodes)
            edges.extend(instances.edges)

        if isinstance(self.mapping, DomainModelApply):
            instances = self.mapping._to_instances_apply(cache)
            nodes.extend(instances.nodes)
            edges.extend(instances.edges)

        if isinstance(self.template, DomainModelApply):
            instances = self.template._to_instances_apply(cache)
            nodes.extend(instances.nodes)
            edges.extend(instances.edges)

        return dm.InstancesApply(dm.NodeApplyList(nodes), dm.EdgeApplyList(edges))


class ScenarioList(TypeList[Scenario]):
    _NODE = Scenario



from __future__ import annotations

from typing import TYPE_CHECKING, ClassVar, Optional, Union

from cognite.client import data_modeling as dm
from pydantic import Field

from cognite.powerops.client._generated.data_classes._core import DomainModel, DomainModelApply, TypeList

if TYPE_CHECKING:
    from cognite.powerops.client._generated.data_classes._output_containers import OutputContainerApply
    from cognite.powerops.client._generated.data_classes._scenario_mappings import ScenarioMappingApply

__all__ = ["ScenarioTemplate", "ScenarioTemplateApply", "ScenarioTemplateList"]


class ScenarioTemplate(DomainModel):
    space: ClassVar[str] = "power-ops"
    base_mapping: Optional[str] = Field(None, alias="baseMapping")
    model: Optional[str] = None
    output_definitions: Optional[str] = Field(None, alias="outputDefinitions")
    shop_files: list[str] = Field([], alias="shopFiles")
    shop_version: Optional[str] = Field(None, alias="shopVersion")
    template_version: Optional[str] = Field(None, alias="templateVersion")
    watercourse: Optional[str] = None


class ScenarioTemplateApply(DomainModelApply):
    space: ClassVar[str] = "power-ops"
    base_mapping: Optional[Union[ScenarioMappingApply, str]] = Field(None, repr=False)
    model: Optional[str] = None
    output_definitions: Optional[Union[OutputContainerApply, str]] = Field(None, repr=False)
    shop_files: list[str] = []
    shop_version: Optional[str] = None
    template_version: Optional[str] = None
    watercourse: Optional[str] = None

    def _to_instances_apply(self, cache: set[str]) -> dm.InstancesApply:
        if self.external_id in cache:
            return dm.InstancesApply(dm.NodeApplyList([]), dm.EdgeApplyList([]))

        sources = []
        properties = {}
        if self.base_mapping is not None:
            properties["baseMapping"] = {
                "space": "power-ops",
                "externalId": self.base_mapping
                if isinstance(self.base_mapping, str)
                else self.base_mapping.external_id,
            }
        if self.model is not None:
            properties["model"] = self.model
        if self.output_definitions is not None:
            properties["outputDefinitions"] = {
                "space": "power-ops",
                "externalId": self.output_definitions
                if isinstance(self.output_definitions, str)
                else self.output_definitions.external_id,
            }
        if self.shop_files is not None:
            properties["shopFiles"] = self.shop_files
        if self.shop_version is not None:
            properties["shopVersion"] = self.shop_version
        if self.template_version is not None:
            properties["templateVersion"] = self.template_version
        if self.watercourse is not None:
            properties["watercourse"] = self.watercourse
        if properties:
            source = dm.NodeOrEdgeData(
                source=dm.ContainerId("power-ops", "ScenarioTemplate"),
                properties=properties,
            )
            sources.append(source)
        if sources:
            this_node = dm.NodeApply(
                space=self.space,
                external_id=self.external_id,
                existing_version=self.existing_version,
                sources=sources,
            )
            nodes = [this_node]
        else:
            nodes = []

        edges = []
        cache.add(self.external_id)

        if isinstance(self.base_mapping, DomainModelApply):
            instances = self.base_mapping._to_instances_apply(cache)
            nodes.extend(instances.nodes)
            edges.extend(instances.edges)

        if isinstance(self.output_definitions, DomainModelApply):
            instances = self.output_definitions._to_instances_apply(cache)
            nodes.extend(instances.nodes)
            edges.extend(instances.edges)

        return dm.InstancesApply(dm.NodeApplyList(nodes), dm.EdgeApplyList(edges))


class ScenarioTemplateList(TypeList[ScenarioTemplate]):
    _NODE = ScenarioTemplate



from __future__ import annotations

from typing import TYPE_CHECKING, ClassVar, Optional, Union

from cognite.client import data_modeling as dm
from pydantic import Field

from cognite.powerops.client._generated.data_classes._core import DomainModel, DomainModelApply, TypeList

if TYPE_CHECKING:
    from cognite.powerops.client._generated.data_classes._bid_matrix_generators import BidMatrixGeneratorApply
    from cognite.powerops.client._generated.data_classes._day_ahead_bids import DayAheadBidApply
    from cognite.powerops.client._generated.data_classes._scenario_mappings import ScenarioMappingApply
    from cognite.powerops.client._generated.data_classes._shop_transformations import ShopTransformationApply

__all__ = ["DayAheadProces", "DayAheadProcesApply", "DayAheadProcesList"]


class DayAheadProces(DomainModel):
    space: ClassVar[str] = "power-ops"
    bid: Optional[str] = None
    bid_matrix_generator_config: list[str] = Field([], alias="bidMatrixGeneratorConfig")
    incremental_mappings: list[str] = []
    name: Optional[str] = None
    shop: Optional[str] = None


class DayAheadProcesApply(DomainModelApply):
    space: ClassVar[str] = "power-ops"
    bid: Optional[Union[DayAheadBidApply, str]] = Field(None, repr=False)
    bid_matrix_generator_config: list[Union[BidMatrixGeneratorApply, str]] = Field(default_factory=list, repr=False)
    incremental_mappings: list[Union[ScenarioMappingApply, str]] = Field(default_factory=list, repr=False)
    name: Optional[str] = None
    shop: Optional[Union[ShopTransformationApply, str]] = Field(None, repr=False)

    def _to_instances_apply(self, cache: set[str]) -> dm.InstancesApply:
        if self.external_id in cache:
            return dm.InstancesApply(dm.NodeApplyList([]), dm.EdgeApplyList([]))

        sources = []
        properties = {}
        if self.bid is not None:
            properties["bid"] = {
                "space": "power-ops",
                "externalId": self.bid if isinstance(self.bid, str) else self.bid.external_id,
            }
        if self.name is not None:
            properties["name"] = self.name
        if self.shop is not None:
            properties["shop"] = {
                "space": "power-ops",
                "externalId": self.shop if isinstance(self.shop, str) else self.shop.external_id,
            }
        if properties:
            source = dm.NodeOrEdgeData(
                source=dm.ContainerId("power-ops", "DayAheadProcess"),
                properties=properties,
            )
            sources.append(source)
        if sources:
            this_node = dm.NodeApply(
                space=self.space,
                external_id=self.external_id,
                existing_version=self.existing_version,
                sources=sources,
            )
            nodes = [this_node]
        else:
            nodes = []

        edges = []
        cache.add(self.external_id)

        for bid_matrix_generator_config in self.bid_matrix_generator_config:
            edge = self._create_bid_matrix_generator_config_edge(bid_matrix_generator_config)
            if edge.external_id not in cache:
                edges.append(edge)
                cache.add(edge.external_id)

            if isinstance(bid_matrix_generator_config, DomainModelApply):
                instances = bid_matrix_generator_config._to_instances_apply(cache)
                nodes.extend(instances.nodes)
                edges.extend(instances.edges)

        for incremental_mapping in self.incremental_mappings:
            edge = self._create_incremental_mapping_edge(incremental_mapping)
            if edge.external_id not in cache:
                edges.append(edge)
                cache.add(edge.external_id)

            if isinstance(incremental_mapping, DomainModelApply):
                instances = incremental_mapping._to_instances_apply(cache)
                nodes.extend(instances.nodes)
                edges.extend(instances.edges)

        if isinstance(self.bid, DomainModelApply):
            instances = self.bid._to_instances_apply(cache)
            nodes.extend(instances.nodes)
            edges.extend(instances.edges)

        if isinstance(self.shop, DomainModelApply):
            instances = self.shop._to_instances_apply(cache)
            nodes.extend(instances.nodes)
            edges.extend(instances.edges)

        return dm.InstancesApply(dm.NodeApplyList(nodes), dm.EdgeApplyList(edges))

    def _create_bid_matrix_generator_config_edge(
        self, bid_matrix_generator_config: Union[str, BidMatrixGeneratorApply]
    ) -> dm.EdgeApply:
        if isinstance(bid_matrix_generator_config, str):
            end_node_ext_id = bid_matrix_generator_config
        elif isinstance(bid_matrix_generator_config, DomainModelApply):
            end_node_ext_id = bid_matrix_generator_config.external_id
        else:
            raise TypeError(f"Expected str or BidMatrixGeneratorApply, got {type(bid_matrix_generator_config)}")

        return dm.EdgeApply(
            space="power-ops",
            external_id=f"{self.external_id}:{end_node_ext_id}",
            type=dm.DirectRelationReference("power-ops", "DayAheadProcess.bidMatrixGeneratorConfig"),
            start_node=dm.DirectRelationReference(self.space, self.external_id),
            end_node=dm.DirectRelationReference("power-ops", end_node_ext_id),
        )

    def _create_incremental_mapping_edge(self, incremental_mapping: Union[str, ScenarioMappingApply]) -> dm.EdgeApply:
        if isinstance(incremental_mapping, str):
            end_node_ext_id = incremental_mapping
        elif isinstance(incremental_mapping, DomainModelApply):
            end_node_ext_id = incremental_mapping.external_id
        else:
            raise TypeError(f"Expected str or ScenarioMappingApply, got {type(incremental_mapping)}")

        return dm.EdgeApply(
            space="power-ops",
            external_id=f"{self.external_id}:{end_node_ext_id}",
            type=dm.DirectRelationReference("power-ops", "DayAheadProcess.incremental_mappings"),
            start_node=dm.DirectRelationReference(self.space, self.external_id),
            end_node=dm.DirectRelationReference("power-ops", end_node_ext_id),
        )


class DayAheadProcesList(TypeList[DayAheadProces]):
    _NODE = DayAheadProces
